#pragma once

#include <opus/opusfile.h>
#include <tgvoip/VoIPController.h>

#include <string>

#include "../libtgvoip/webrtc_dsp/common_audio/wav_file.h"
#include "utils.hpp"

struct TgVoIPControllerWrapper {
    TgVoIPControllerWrapper(const std::string &reflector_host_port,
                            const std::string &tag_hex,
                            const std::string &encryption_key_hex,
                            bool               is_outgoing,
                            const std::string  input_file      = "",
                            const std::string  output_file     = "",
                            const std::string &proxy_host_port = "")
    {
        if (!proxy_host_port.empty()) {
            auto [host, port] = parse_host_port(proxy_host_port);
            controller_.SetProxy(tgvoip::PROXY_SOCKS5, host, port, "", "");
        }

        if (!reflector_host_port.empty()) {
            auto [host, port] = parse_host_port(reflector_host_port);
            tgvoip::IPv4Address host_ip{host};
            tgvoip::IPv6Address emptyV6;
            auto                tag = hex2bin(tag_hex);
            tgvoip::Endpoint    ep{1, port, host_ip, emptyV6, tgvoip::Endpoint::Type::UDP_RELAY, tag.data()};
            controller_.SetRemoteEndpoints({ep}, false, 76);
        }

        if (!encryption_key_hex.empty()) {
            auto encryption_key = hex2bin(encryption_key_hex);
            assert(encryption_key.size() == 256);  // a bit insecure API, imho
            controller_.SetEncryptionKey(reinterpret_cast<char *>(encryption_key.data()), is_outgoing);
        }

        tgvoip::VoIPController::Config controller_config{};
        controller_config.enableNS = controller_config.enableAEC = controller_config.enableAGC = false;

        controller_config.enableCallUpgrade = false;
        controller_config.initTimeout       = 5.0;
        controller_config.recvTimeout       = 3.0;

        controller_.SetConfig(controller_config);

        std::string real_input_file = input_file;
        if (real_input_file.empty()) {
            real_input_file = is_outgoing ? "samples-test/input-caller.wav" : "samples-test/input-callee.wav";
        }

        std::string real_output_file = output_file;
        if (real_output_file.empty()) {
            real_output_file = is_outgoing ? "samples-test/output-caller.wav" : "samples-test/output-callee.wav";
        }

        //        reader_ = new webrtc::WavReader(real_input_file);
        writer_ = new webrtc::WavWriter(real_output_file, 48000, 1);

        //        controller_.SetAudioDataCallbacks(
        //            [this](int16_t *data, size_t len) {
        //                if (reader_ != nullptr) reader_->ReadSamples(len, data);
        //            },
        //            [this](int16_t *data, size_t len) {
        //                if (writer_ != nullptr) writer_->WriteSamples(data, len);
        //            });

        in_file_            = op_open_file(real_input_file.c_str(), nullptr);
        int in_num_channels = op_channel_count(in_file_, -1);

        out_file_ = op_open_file(real_output_file.c_str(), nullptr);

        controller_.SetAudioDataCallbacks(
            [this, in_num_channels](int16_t *data, size_t len) {
                int samples_read = op_read(in_file_, data, len * sizeof(int16_t) * in_num_channels, nullptr);
                if (samples_read != 0 && samples_read != len) {
                    fprintf(stderr, "!!!!!! ERROR: %d %d\n", samples_read, len);
                } else if (samples_read == len) {
                    //                    fprintf(stderr, "!!!!!! OK %d %d\n", samples_read, len);
                }
            },
            [this](int16_t *data, size_t len) {
                if (writer_ != nullptr) writer_->WriteSamples(data, len);
            });
    }

    void start() { controller_.Start(); }
    void connect() { controller_.Connect(); }

    void setAudioDataCallbacks(std::function<void(int16_t *, size_t)> input, std::function<void(int16_t *, size_t)> output)
    {
        controller_.SetAudioDataCallbacks(input, output);
    }

    std::string getDebugLog() { return controller_.GetDebugLog(); }

    ~TgVoIPControllerWrapper()
    {
        //        delete reader_;
        delete writer_;
        //        reader_ = nullptr;
        writer_ = nullptr;
        if (in_file_) op_free(in_file_);
        if (out_file_) op_free(out_file_);

        controller_.Stop();
    }

private:
    tgvoip::VoIPController controller_;

    OggOpusFile *in_file_;
    OggOpusFile *out_file_;
    //    webrtc::WavReader *reader_;
    webrtc::WavWriter *writer_;
};
